#lang scheme
(require "cub-exp.ss"
         (prefix-in e-e: "../eval/eval-exp.ss")
         redex/reduction-semantics)

(provide tree-exp-red tree-exp+ec Eval MovePC)

(define-extended-language tree-exp+ec
  cub-exp
  (flatten-S (move (mem flatten-E) tree-exp)
             (move (mem pure-exp) flatten-E)
             (move temp flatten-E)
             (texp flatten-E)
             (jump flatten-E label ...)
             (cjump relop flatten-E tree-exp
                    label label)
             (cjump relop pure-exp flatten-E
                    label label))
  (flatten-E hole
             (in-hole flatten-E1 flatten-E))
  (flatten-E1 (eseq flatten-S tree-exp)
              (biop hole tree-exp)
              (biop pure-exp hole)
              (mem hole)
              (call fn hole))
  
  (S ((any any) ...))
  (int (side-condition number_1 (and (integer? (term number_1))
                                     (exact? (term number_1)))))
  (bool #t #f)
  (pure-tree-v int label))
  
(define tree-exp-red
  (reduction-relation 
   tree-exp+ec
   (--> (S
         tree-stm_before ...
         pc
         (seq tree-stm_1 ...)
         tree-stm_after ...)
        (S
         tree-stm_before ...
         pc
         tree-stm_1 ...
         tree-stm_after ...)
        flatten-seq)
   
   (--> (S
         tree-stm_before ...
         pc
         (in-hole flatten-S (eseq tree-stm tree-exp))
         tree-stm_after ...)
        (S
         tree-stm_before ...
         pc 
         tree-stm
         (in-hole flatten-S tree-exp) 
         tree-stm_after ...)
        flatten-eseq)
   
   (--> (S
         tree-stm_before ...
         pc
         (in-hole flatten-S (in-hole flatten-E1 (call fn pure-exp)))
         tree-stm_after ...)
        (S
         tree-stm_before ...
         pc 
         (move r:temp (call fn pure-exp))
         (in-hole flatten-S (in-hole flatten-E1 r:temp)) 
         tree-stm_after ...)
        flatten-call
        (fresh r:temp))
   
   (--> (S
         tree-stm_before ...
         pc
         (move temp pure-exp_1)
         tree-stm_after ...)
        ((update S temp (Eval S pure-exp_1))
         tree-stm_before ...
         (move temp pure-exp_1)
         pc
         tree-stm_after ...)
        move-temp-exp)
   
   (--> (S
         
         
         tree-stm_before ...
         pc
         (move (mem pure-exp_1)
               pure-exp_2)
         tree-stm_after ...)
        ((update S
                 (Eval S pure-exp_1)
                 (Eval S pure-exp_2))
         tree-stm_before ...
         (move (mem pure-exp_1)
               pure-exp_2)
         pc
         tree-stm_after ...)
        move-mem-exp)
   
   
   (--> (S
         tree-stm_before ...
         pc
         (move (mem pure-exp_1)
               (call fn pure-exp_2))
         tree-stm_after ...)
        (S
         tree-stm_before ...
         pc
         (move r:temp (call fn pure-exp_2))
         (move (mem pure-exp_1) r:temp)
         tree-stm_after ...)
        (fresh r:temp)
        move-mem-call)
   
   (--> (S
         tree-stm_before ...
         pc
         (move temp (call "allocate" pure-exp))
         tree-stm_after ...)
        ((alloc S temp (Eval S pure-exp))
         tree-stm_before ...
         (move temp (call "allocate" pure-exp))
         pc
         tree-stm_after ...)
        move-temp-alloc)
   
   (--> (S
         tree-stm_before ...
         pc
         (move temp (call fn pure-exp))
         tree-stm_after ...)
        ((update S temp 0)
         tree-stm_before ...
         (move temp (call fn pure-exp))
         pc
         tree-stm_after ...)
        move-temp-fn
        (side-condition (not (equal? (term fn) (term "allocate")))))
   
   (--> (S
         tree-stm_before ...
         pc
         (jump pure-exp label ...)
         tree-stm_after ...)
        (MovePC (Eval S pure-exp) 
                (S
                 tree-stm_before ...
                 (jump pure-exp label ...)
                 tree-stm_after ...))
        jump)
   
   (--> (S
         tree-stm_before ...
         pc
         (cjump biop
                pure-exp_1 pure-exp_2
                label_n label_0)
         tree-stm_after ...)
        (MovePC label_n
                (S
                 tree-stm_before ...
                 (cjump biop
                        pure-exp_1 pure-exp_2
                        label_n label_0)
                 tree-stm_after ...))
        (side-condition (term (Nonzero? (Eval S (biop pure-exp_1 pure-exp_2)))))
        cjump-true)
   
   (--> (S
         tree-stm_before ...
         pc
         (cjump biop 
                pure-exp_1 pure-exp_2
                label_n label_0)
         tree-stm_after ...)
        (MovePC label_0
                (S
                 tree-stm_before ...
                 (cjump biop
                        pure-exp_1 pure-exp_2
                        label_n label_0)
                 tree-stm_after ...))
        (side-condition (term (Zero? (Eval S (biop pure-exp_1 pure-exp_2)))))
        cjump-false)
   
   (--> (S
         tree-stm_before ...
         pc
         label
         tree-stm_after ...)
        (S
         tree-stm_before ...
         label
         pc
         tree-stm_after ...)
        label)
   
   (--> (S
         tree-stm_before ...
         pc
         (texp pure-exp)
         tree-stm_after ...)
        (S
         tree-stm_before ...
         (texp pure-exp)
         pc
         tree-stm_after ...)
        texp)))
        
;; moves the program counter to the given label
(define-metafunction tree-exp+ec
  MovePC : label (S tree-stm ...) -> (S tree-stm ... pc tree-stm ...)
  [(MovePC label (S tree-stm_before ... label tree-stm_after ...))
   (S tree-stm_before ... label pc tree-stm_after ...)])

;; allocate & initialize 'num' new bytes in the store and store the result in 'temp'
(define-metafunction tree-exp+ec
  alloc : S temp num -> S
  [(alloc ((any_lhs any_rhs) ...) temp num) 
   (init (update ((any_lhs any_rhs) ...) temp new-loc) new-loc num)
   (where new-loc ,(apply max 4 (filter number? (term (any_lhs ...)))))])
   
;; initialize 'num_count' bytes starting at location 'num_count'
(define-metafunction tree-exp+ec
  init : S num num -> S
  [(init S num_loc 0) 
   S]
  [(init S num_loc num_count)
   (init (update S num_loc 0) ,(+ (term num_loc) 4) ,(- (term num_count) 1))])

;; evaluate a pure expression with respect to the store S
(define-metafunction tree-exp+ec
  Eval : S pure-exp -> pure-tree-v
  [(Eval S num) num]
  [(Eval S label) label]
  [(Eval S temp) (lookup S temp)]
  [(Eval S (mem 
            pure-exp))
   (lookup S (Eval S
                   pure-exp))]
  [(Eval S (biop 
            pure-exp_1
            pure-exp_2)) 
   (δ biop 
      (Eval S pure-exp_1)
      (Eval S pure-exp_2))])

;; lookup something in the store
(define-metafunction tree-exp+ec
  lookup : S any -> any
  [(lookup (any_1 ... (any_2 any_3) any_4 ...) any_2) any_3])

;; update the store
(define-metafunction tree-exp+ec
  update : S any any -> S
  [(update (any_1 ... (any_2 any_3) any_4 ...) any_2 any_5)
   (any_1 ... (any_2 any_5) any_4 ...)]
  [(update (any_1 ...) any_2 any_5)
   (any_1 ... (any_2 any_5))])

;; a hack: use the delta function from elsewhere here
(define-metafunction tree-exp+ec
  δ : biop any any -> any
  [(δ biop any_1 any_2) ,(e-e:δ (term biop) (term any_1) (term any_2))])

(define-metafunction tree-exp+ec
  Zero? : num -> bool
  [(Zero? 0) #t]
  [(Zero? num) #f])

(define-metafunction tree-exp+ec
  Nonzero? : num -> bool
  [(Nonzero? 0) #f]
  [(Nonzero? num) #t])

